import java.util.Random;

import percolation.Cell;
import percolation.Grid;
import percolation.State;
import percolation.StateOperation;

public class Test {

  private static void test1a() {
    for (State from : State.values()) {
      for (State to : State.values())
        System.out.printf(" %.1B", StateOperation.allowTransition(from, to));
      System.out.println();
    }
  }

  private static void test1b() {
    for (State state1 : State.values()) {
      for (State state2 : State.values())
        System.out.printf(" %12s", StateOperation.fusion(state1, state2));
      System.out.println();
    }
  }

  private static void assertEnabled() {
    try {
      assert false;
      System.err.println("Il manque l'option -ea");
      System.exit(1);
    } catch (AssertionError e) {
    }
  }

  private static void assertSetState(Cell c1, Cell c2, State s1, State s2) {
    c1.setState(s1);
    c2.setState(s2);
    assert c1.getState() == s1 : "getState after setState";
    assert c2.getState() == s2 : "getState after setState";
  }

  private static void assertStateException(Cell c, State newState) {
    try {
      c.setState(newState);
    } catch (IllegalArgumentException e) {
      return; // OK
    }
    throw new AssertionError(
        "setState should throw an IllegalArgumentException in such a case");
  }

  private static void test2() {
    assertEnabled();
    Cell c1 = new Cell();
    Cell c2 = new Cell();
    assert c1.getState() == State.BLOCKED;
    assert c2.getState() == State.BLOCKED;
    assertSetState(c1, c2, State.OPEN, State.BLOCKED);
    assertSetState(c1, c2, State.INPUT, State.OUTPUT);
    assertStateException(c1, State.OUTPUT);
    assertStateException(c2, State.INPUT);
    c1.reset();
    assert c1.getState() == State.BLOCKED : "after reset";
    assert c2.getState() == State.OUTPUT : "after reset";
    c2.reset();
    assert c2.getState() == State.BLOCKED : "after reset";
    System.out.println("Test 2 OK.");
  }

  private static void assertEquals(State s, Cell... cs) {
    for (Cell c : cs)
      assert c.getState() == s : "shared state after union";
  }

  private static void assertUnion(Cell c1, Cell c2, State s1, State s2, State s) {
    c1.setState(s1);
    c2.setState(s2);
    Cell.union(c1, c2);
    assertEquals(s, c1, c2);
  }

  private static void assertReset(Cell... cs) {
    for (Cell c : cs) {
      c.reset();
      assert c.getState() == State.BLOCKED : "getState after reset";
    }
  }

  private static void test3a() {
    assertEnabled();
    Cell c1 = new Cell();
    Cell c2 = new Cell();
    assertUnion(c1, c2, State.OPEN, State.INPUT, State.INPUT);
    c1.setState(State.PERCOLATING);
    assertEquals(State.PERCOLATING, c1, c2);
    assertReset(c1, c2);
    System.out.println("Test 3a OK.");
  }

  private static void test3b() {
    assertEnabled();
    Cell c1 = new Cell();
    Cell c2 = new Cell();
    assertUnion(c1, c2, State.OPEN, State.INPUT, State.INPUT);
    Cell c3 = new Cell();
    Cell c4 = new Cell();
    assertUnion(c3, c4, State.INPUT, State.OPEN, State.INPUT);
    Cell.union(c2, c3);
    c1.setState(State.PERCOLATING);
    assertEquals(State.PERCOLATING, c1, c2, c3, c4);
    assertReset(c1, c3, c4, c2);
    System.out.println("Test 3b OK.");
  }

  private static void test3c() {
    assertEnabled();
    Cell c1 = new Cell();
    c1.setState(State.OPEN);
    Cell c2 = new Cell();
    c2.setState(State.INPUT);
    Cell c3 = new Cell();
    c3.setState(State.OUTPUT);
    Cell c4 = new Cell();
    c4.setState(State.OPEN);
    Cell.union(c1, c2);
    Cell.union(c3, c4);
    Cell.union(c4, c1);
    Cell.union(c2, c3);
    assertEquals(State.PERCOLATING, c1, c2, c3, c4);
    assertReset(c1, c3, c4, c2);
    System.out.println("Test 3c OK.");
  }

  private static Cell[] makeArray(int n) {
    Cell[] t = new Cell[n];
    for (int i = 0; i < n; ++i)
      t[i] = new Cell();
    return t;
  }

  private static void test3d(int n) {
    assertEnabled();
    Cell[] t = makeArray(n);
    System.gc();
    Cell c = t[0];
    for (int i = 1; i < t.length; ++i)
      Cell.union(c, t[i]);
    t[0].setState(State.OPEN);
    assertEquals(State.OPEN, t);
    assertReset(t);
    State[] values = State.values();
    for (int i = 0; i < t.length; ++i)
      t[i].setState(values[i % values.length]);
    for (int i = 0; i < t.length; ++i)
      assert t[i].getState() == values[i % values.length];

    t = makeArray(n);
    System.gc();
    long start = System.currentTimeMillis();
    c = t[0];
    for (int i = 1; i < t.length; ++i)
      Cell.union(c, t[i]);
    System.out.println("build time : " + (System.currentTimeMillis() - start)
        + " ms.");
    t[0].setState(State.OPEN);
    start = System.currentTimeMillis();
    assertEquals(State.OPEN, t);
    System.out.println("query time : " + (System.currentTimeMillis() - start)
        + " ms.");

    t = makeArray(n);
    System.gc();
    start = System.currentTimeMillis();
    c = t[0];
    for (int i = 1; i < t.length; ++i)
      Cell.union(t[i], c);
    System.out.println("build time : " + (System.currentTimeMillis() - start)
        + " ms.");
    t[0].setState(State.OPEN);
    start = System.currentTimeMillis();
    assertEquals(State.OPEN, t);
    System.out.println("query time : " + (System.currentTimeMillis() - start)
        + " ms.");
  }

  private static Grid build(int n) {
    Grid g = new Grid(n);
    g.getCell(0, 0).setState(State.INPUT);
    for (int i = 1; i < g.size() - 1; ++i)
      g.getCell(i, g.size() - i - 1).setState(State.OPEN);
    g.getCell(g.size() - 1, g.size() - 1).setState(State.OUTPUT);
    return g;
  }

  private static void test4() {
    new DisplayFrame(build(3));
    Grid g = build(3);
    new DisplayFrame(g);
    g.reset();
    new DisplayFrame(build(4));
  }

  private static void test5a() {
    Grid g = new Grid(5);
    g.open(0, 0);
    g.open(1, 1);
    g.open(2, 2);
    g.open(2, 1);
    g.open(3, 1);
    g.open(2, 0);
    g.open(0, 4);
    g.open(0, 3);
    g.open(4, 4);
    g.open(1, 3);
    g.open(1, 2);
    new DisplayFrame(g);
  }

  private static void test5b(int n) {
    Random rand = new Random(12345);
    Grid g = new Grid(20);
    for (int k = 0; k < n; ++k) {
      int i = rand.nextInt(g.size());
      int j = rand.nextInt(g.size());
      if (g.getCell(i, j).getState() == State.BLOCKED)
        g.open(i, j);
    }
    new DisplayFrame(g);
  }

  public static void main(String[] args) {
    test1a();
    test1b();
    // test2();
    // test3a();
    // test3b();
    // test3c();
    // test3d(1000000);
    // test4();
    // test5a();
    // test5b(358);
    // test5b(359);
  }

}
